#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['keggtools']

package_data = \
{'': ['*']}

install_requires = \
['pydot >= 1.4.2', 'requests >= 2.27.1', 'scipy >= 1.5.4']

extras_require = \
{'doc': ['Sphinx', 'sphinx-rtd-theme'],
 'test': ['pylint',
          'pytest',
          'coverage',
          'mypy',
          'types-requests',
          'responses',
          'pandas']}

setup(name='keggtools',
      version='1.0.0',
      description='Enrichment analysis and visualisation toolkit for KEGG pathways',
      author=None,
      author_email='harryhaller001 <harryhaller001@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
